/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998, 1999  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    slouken@libsdl.org
*/

#include "SDL_direct3dvideo.h"
#include "SDL_direct3dtexturegrid.h"
#include "SDL_blit.h"
#include "SDL_wingl_c.h"

#if 0
#define SDL_D3D_DEVICE_TYPE D3DDEVTYPE_REF
#else
#define SDL_D3D_DEVICE_TYPE D3DDEVTYPE_HAL
#endif

/* Transformed vertex with 1 set of texture coordinates */
const DWORD tri_fvf = D3DFVF_XYZRHW | D3DFVF_TEX1 | D3DFVF_DIFFUSE;

/* Initialization/Query functions */
static int DIRECT3D_CreateWindow(_THIS);
static int DIRECT3D_VideoInit(_THIS, SDL_PixelFormat *vformat);
static SDL_Rect **DIRECT3D_ListModes(_THIS, SDL_PixelFormat *foramt, UINT32 flags);
static SDL_Surface *DIRECT3D_SetVideoMode(_THIS, SDL_Surface *current, int width, int height, int bpp, Uint32 flags);
static int DIRECT3D_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors);
static int DIRECT3D_SetGammaRamp(_THIS, Uint16 *ramp);
static int DIRECT3D_GetGammaRamp(_THIS, Uint16 *ramp);
static void DIRECT3D_VideoQuit(_THIS);

/* Hardware surface functions */
static int DIRECT3D_AllocHWSurface(_THIS, SDL_Surface *surface);
static int DIRECT3D_CheckHWBlit(_THIS, SDL_Surface *src, SDL_Surface *dst);
static int DIRECT3D_HWAccelBlit(SDL_Surface *src, SDL_Rect *srcrect, SDL_Surface *dst, SDL_Rect *dstrect);
static int DIRECT3D_FillHWRect(_THIS, SDL_Surface *dst, SDL_Rect *dstrect, Uint32 color);
static int DIRECT3D_SetHWColorKey(_THIS, SDL_Surface *surface, Uint32 key);
static int DIRECT3D_SetHWAlpha(_THIS, SDL_Surface *surface, Uint8 alpha);
static int DIRECT3D_LockHWSurface(_THIS, SDL_Surface *surface);
static void DIRECT3D_UnlockHWSurface(_THIS, SDL_Surface *surface);
static int DIRECT3D_FlipHWSurface(_THIS, SDL_Surface *surface);
static void DIRECT3D_FreeHWSurface(_THIS, SDL_Surface *surface);
static void DIRECT3D_UpdateRects(_THIS, int numrects, SDL_Rect *rects);

/* Windows message handling functions */
static void DIRECT3D_RealizePalette(_THIS);
static void DIRECT3D_PaletteChanged(_THIS, HWND window);
static void DIRECT3D_WinPAINT(_THIS, HDC hdc);
static LONG DIRECT3D_HandleMessage(_THIS, HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

/* WinDIB imports */
extern int DIB_CreateWindow(_THIS);
extern void DIB_WinPAINT(_THIS, HDC hdc);
extern void DIB_PumpEvents(_THIS);

/* misc imports */
extern SDL_PixelFormat *SDL_AllocFormat(int bpp, Uint32 Rmask, Uint32 Gmask, Uint32 Bmask, Uint32 Amask);

/* misc functions */
LONG CALLBACK WinMessage(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static HRESULT set_device_states(_THIS, int mode, SDLD3D_BlendMode blend_mode);
static HRESULT begin_rendering(_THIS);
static HRESULT stop_rendering(_THIS);

/* device states, as recognized by set_device_states */
#define SDLD3D_DEVMODE_DRAWTEXTURES		1
#define SDLD3D_DEVMODE_DRAWSOLIDCOLORS	2

/* The global reference to the Direct3D library. */
HINSTANCE Direct3D_DLL = NULL;

/* Global screen flags.  See SDL_direct3dext.h for more info on these. */
SDLD3D_ScreenFlags g_ScreenFlags = 0;

/* Direct3D function pointers. */
static LPDIRECT3D9 (WINAPI *Direct3DCreate)(UINT SDKVersion);



/* Unload the Direct3D subsystem. */
static void DIRECT3D_Unload(void)
{
	if ( Direct3D_DLL != NULL ) {
		FreeLibrary(Direct3D_DLL);
		Direct3DCreate = NULL;
		Direct3D_DLL = NULL;
	}
}

/* Load the Direct3D subsystem. */
static int DIRECT3D_Load(void)
{
	int status;

	/* If Direct3D has already been loaded, it will be reloaded. */
	DIRECT3D_Unload();

	/* Try to load the Direct3D library. */
	Direct3D_DLL = LoadLibrary(TEXT("D3D9.DLL"));
	if ( Direct3D_DLL != NULL )
	{
		/* Try to retrieve the Direct3D initialization function. */
		Direct3DCreate = (void *)GetProcAddress(Direct3D_DLL, TEXT("Direct3DCreate9"));
	}

	if ( Direct3D_DLL && Direct3DCreate ) {
		/* All functions were loaded ok.  Return a success code of '0'. */
		status = 0;
	}
	else
	{
		/* An error occurred.  Clean up and return a failure code of '-1'. */
		DIRECT3D_Unload();
		status = -1;
	}

	return status;
}

/* Check if Direct3D support is available. */
static int DIRECT3D_Available(void)
{
	int d3d_ok = 0;

	if ( DIRECT3D_Load() == 0 )
	{
		LPDIRECT3D9 D3D;

		/* Direct3D was loaded successfully.  Now try calling 'Direct3DCreate'. */
		D3D = Direct3DCreate(D3D_SDK_VERSION);
		if ( D3D != NULL )
		{
			/* 'Direct3DCreate' succeeded.  Destroy the Direct3D object and
			   make sure a success value of '1' is returned. */
			IDirect3D9_Release(D3D);
			d3d_ok = 1;
		}
	}

	/* Clean up. */
	DIRECT3D_Unload();

	return d3d_ok;
}

static void DIRECT3D_DeleteDevice(_THIS)
{
	/* Destroy the Direct3D Device */
	IDirect3DDevice9_Release(d3d9_device);

	/* Destroy the Direct3D Object */
	IDirect3D9_Release(d3d9_object);

	/* Free the Direct3D functions + library */
	DIRECT3D_Unload();

	/* Free memory associated with the SDL_PrivateVideoData struct */
	if ( this ) {
		if ( this->gl_data ) {
			free(this->gl_data);
		}

		if ( this->hidden ) {
			free(this->hidden);
		}

		free(this);
	}
}

static SDL_VideoDevice *DIRECT3D_CreateDevice(int devindex)
{
	SDL_VideoDevice *device;

	/* Load Direct3D */
	if ( DIRECT3D_Load() < 0 ) {
		return NULL;
	}

	/* Initialize all variables that we clean on shutdown. */
	device = (SDL_VideoDevice *)malloc(sizeof(SDL_VideoDevice));
	if ( device ) {
		memset(device, 0, (sizeof *device));
		device->hidden = (struct SDL_PrivateVideoData *)malloc((sizeof *device->hidden));
		device->gl_data = (struct SDL_PrivateGLData *)malloc((sizeof *device->gl_data));
	}
	if ( (device == NULL) || (device->hidden == NULL) ) {
		SDL_OutOfMemory();
		DIRECT3D_DeleteDevice(device);
		return NULL;
	}
	memset(device->hidden, 0, (sizeof *device->hidden));
	memset(device->gl_data, 0, (sizeof *device->gl_data));
	memset(&(device->hidden->__d3d_presentation_parameters), 0, sizeof(device->hidden->__d3d_presentation_parameters));

	/* Set the function pointers. */
	device->VideoInit = DIRECT3D_VideoInit;
	device->ListModes = DIRECT3D_ListModes;
	device->SetVideoMode = DIRECT3D_SetVideoMode;
	device->UpdateMouse = WIN_UpdateMouse;
	device->CreateYUVOverlay = NULL;
	device->SetColors = DIRECT3D_SetColors;
	device->UpdateRects = NULL;
	device->VideoQuit = DIRECT3D_VideoQuit;
	device->AllocHWSurface = DIRECT3D_AllocHWSurface;
	device->CheckHWBlit = DIRECT3D_CheckHWBlit;
	device->FillHWRect = DIRECT3D_FillHWRect;
	device->SetHWColorKey = DIRECT3D_SetHWColorKey;
	device->SetHWAlpha = DIRECT3D_SetHWAlpha;
	device->LockHWSurface = DIRECT3D_LockHWSurface;
	device->UnlockHWSurface = DIRECT3D_UnlockHWSurface;
	device->FlipHWSurface = DIRECT3D_FlipHWSurface;
	device->FreeHWSurface = DIRECT3D_FreeHWSurface;
	device->UpdateRects = DIRECT3D_UpdateRects;
	device->SetGammaRamp = DIRECT3D_SetGammaRamp;
	device->GetGammaRamp = DIRECT3D_GetGammaRamp;
#ifdef HAVE_OPENGL
	device->GL_LoadLibrary = NULL;
	device->GL_GetProcAddress = NULL;
	device->GL_GetAttribute = NULL;
	device->GL_MakeCurrent = NULL;
	device->GL_SwapBuffers = NULL;
#endif
	device->SetCaption = WIN_SetWMCaption;
	device->SetIcon = WIN_SetWMIcon;
	device->IconifyWindow = WIN_IconifyWindow;
	device->GrabInput = WIN_GrabInput;
	device->GetWMInfo = WIN_GetWMInfo;
	device->FreeWMCursor = WIN_FreeWMCursor;
	device->CreateWMCursor = WIN_CreateWMCursor;
	device->ShowWMCursor = WIN_ShowWMCursor;
	device->WarpWMCursor = WIN_WarpWMCursor;
	device->CheckMouseMode = WIN_CheckMouseMode;
	device->InitOSKeymap = DIB_InitOSKeymap;
	device->PumpEvents = DIB_PumpEvents;

	/* Set up the windows message handling functions */
	WIN_RealizePalette = DIRECT3D_RealizePalette;
	WIN_PaletteChanged = DIRECT3D_PaletteChanged;
	WIN_WinPAINT = DIRECT3D_WinPAINT;
	HandleMessage = DIRECT3D_HandleMessage;

	device->free = DIRECT3D_DeleteDevice;

	/* We're finally ready. */
	return device;
}


VideoBootStrap DIRECT3D_bootstrap = {
	"direct3d", "Windows Direct3D",
	DIRECT3D_Available, DIRECT3D_CreateDevice
};

const char * DescribeLastWinError()
{
	static char buf[512];

	if (!FormatMessage( 
		FORMAT_MESSAGE_FROM_SYSTEM | 
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		GetLastError(),
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
		(LPTSTR) buf,
		512,
		NULL ))
	{
		// Handle the error.
		return "Could not retrieve error info.";
	}

	return buf;
}

const char * DescribeLastDXError()
{
	return DXGetErrorString9(GetLastError());
}

LRESULT CALLBACK DIRECT3D_WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
	case WM_GETMINMAXINFO:
		// FIXME: Passing WM_GETMINMAXINFO will cause the Direct3D driver to fail.  Why?
		return DefWindowProc(hWnd, message, wParam, lParam);

	default:
		return WinMessage(hWnd, message, wParam, lParam);
	}

	return 0;
}

int DIRECT3D_CreateWindow(_THIS)
{
	int result = DIB_CreateWindow(this);

	if ( result == 0 ) {
		SetWindowLong(SDL_Window, GWL_WNDPROC, (LONG)DIRECT3D_WndProc);
		ShowWindow(SDL_Window, SW_HIDE);
		UpdateWindow(SDL_Window);
	}

	return result;
}

void DIRECT3D_DestroyWindow(_THIS)
{
	DIB_DestroyWindow(this);
}

static int DIRECT3D_UpdateVideoInfo(_THIS)
{
	/* Size of video memory, in bytes */
	UINT vid_mem_size;

	VERIFY_FCALL(IDirect3D9_GetDeviceCaps, (
		d3d9_object,
		/* Adapter */
		D3DADAPTER_DEFAULT,
		/* DeviceType */
		SDL_D3D_DEVICE_TYPE,
		/* pCaps (the output) */
		&d3d9_device_caps),
		/* return -1 on error */
		return -1);

	//pow2 = d3d9_device_caps.TextureCaps & D3DPTEXTURECAPS_POW2;

	/* FIXME, davidl: Implement this function.  -ALL- current filled in values
	are temporary, and were placed here just to get things running. */

	/* Fill in our hardware acceleration capabilities */
	this->info.hw_available = 1;
	this->info.blit_hw = 1;
	//this->info.blit_hw_CC = 1;
	this->info.blit_hw_A = 1;
	this->info.blit_fill = 1;

	this->displayformatalphapixel = SDL_AllocFormat(32, 0x00FF0000, 0x0000FF00, 0x000000FF, 0xFF000000);
	if ( this->displayformatalphapixel == NULL )
	{
		SDL_OutOfMemory();
		return -1;
	}

	/* Find out how much video memory is available.
	   NOTE: This value is only an estimate!  See MSDN's documentation on
	   IDirect3DDevice9::GetAvailableTextureMem for more details. */
	if ( d3d9_device ) {
		vid_mem_size = IDirect3DDevice9_GetAvailableTextureMem(d3d9_device);
		this->info.video_mem = vid_mem_size / 1024;
	}
	else
	{
		this->info.video_mem = 0;
	}

	return 0;
}

int DIRECT3D_VideoInit(_THIS, SDL_PixelFormat *vformat)
{
	int i;
	HDC hdc;

	/* HACK, davidl: Enable 16 bit color at 640x480 */
	static SDL_Rect rect;

	/* Create the Wwindow */
	if ( DIRECT3D_CreateWindow(this) < 0 ) {
		return -1;
	}

	/* Create the Direct3d object */
	this->hidden->__d3d9_object = Direct3DCreate(D3D_SDK_VERSION);
	if ( this->hidden->__d3d9_object == NULL )
	{
		SDL_SetError("Direct3DCreate failed");
		return -1;
	}

	/* Determine the screen depth */
	hdc = GetDC(SDL_Window);
	vformat->BitsPerPixel = GetDeviceCaps(hdc, PLANES) * GetDeviceCaps(hdc, BITSPIXEL);
	ReleaseDC(SDL_Window, hdc);

	/* Enumerate the available fullscreen modes */
	/* FIXME, davidl: Implement mode scanning. */

	for ( i=0; i<NUM_MODELISTS; ++i ) {

		/* HACK, davidl: Enable 16 bit color at 640x480 */
		if ( i == 1 ) {
			SDL_nummodes[i] = 1;
		} else {
			SDL_nummodes[i] = 0;
		}

		SDL_modelist[i] = (SDL_Rect **)malloc((SDL_nummodes[i]+1) * sizeof(SDL_Rect *));
		if ( SDL_modelist[i] == NULL ) {
			SDL_OutOfMemory();
			return -1;
		}

		/* HACK, davidl: Enable 16 bit color at 40x480 */
		if ( i == 1 ) {
			SDL_modelist[i][1] = NULL;
		} else {
			SDL_modelist[i][0] = NULL;
		}
	}

	/* HACK, davidl: Enable 16 bit color at 640x480 */
	SDL_modelist[1][0] = &rect;
	SDL_modelist[1][0]->x = 0;
	SDL_modelist[1][0]->y = 0;
	SDL_modelist[1][0]->w = 640;
	SDL_modelist[1][0]->h = 480;

	/* Fill in some window manager capabilities */
	this->info.wm_available = 1;

	/* Fill in some video hardware capabilities */
	DIRECT3D_UpdateVideoInfo(this);

	return 0;
}

SDL_Rect **DIRECT3D_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags)
{
	int bpp;

	bpp = format->BitsPerPixel;
	if ( (flags & SDL_FULLSCREEN) == SDL_FULLSCREEN ) {
		/* FIXME, davidl: Add support for fullscreen */
		return (SDL_Rect **) -1;
	} else {
		return (SDL_Rect **) -1;
		/*
		if ( this->screen->format->BitsPerPixel = bpp ) {
			return (SDL_Rect **) -1;
		} else {
			return (SDL_Rect **) 0;
		}
		/**/
	}
	return NULL;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	return DefWindowProc(hWnd, message, wParam, lParam);
}

SDL_Surface *DIRECT3D_SetVideoMode(_THIS, SDL_Surface *current,
								   int width, int height, int bpp, Uint32 flags)
{
	SDL_Surface *video;
	//HRESULT result;
	//DWORD style;
	const DWORD directstyle = (WS_POPUP);
	const DWORD windowstyle = (WS_OVERLAPPED|WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX);
	const DWORD resizestyle = (WS_THICKFRAME|WS_MAXIMIZEBOX);
	int windowX, windowY;
	LONG style;
	D3DFORMAT format;
	UINT presentation_interval;
	//IDirect3DSurface9 *backBuffer;
	struct private_hwdata *surfaceData;

	/* Clean up any previous buffers */
	if ( current->hwdata ) {
		this->FreeHWSurface(this, current);
		current->hwdata = NULL;
	}
	/* FIXME, davidl: more cleanup has to be done here. */

	/* Calculate the window's new size, given a client area of (width, height). */
	style = GetWindowLong(SDL_Window, GWL_STYLE);

	if ( (flags & SDL_FULLSCREEN) == SDL_FULLSCREEN ) {
		//style &= ~windowstyle;
		//style |= directstyle;
		style = WS_POPUP;
	} else {
		if ( flags & SDL_NOFRAME ) {
			style &= ~windowstyle;
			style |= directstyle;
		} else {
			style &= ~directstyle;
			style |= windowstyle;
			if ( flags & SDL_RESIZABLE ) {
				style |= resizestyle;
			}
		}
	}

	/* DJM: Don't piss of anyone who has setup his own window */
	if ( SDL_windowid == NULL )
		SetWindowLong(SDL_Window, GWL_STYLE, style);

	/* Set the display mode, if we are in fullscreen mode */
	if ( (flags & SDL_FULLSCREEN) == SDL_FULLSCREEN ) {
		RECT bounds;

		/* Cover up desktop during mode change */
		//SDL_resizing = 1;
		bounds.left = 0;
		bounds.top = 0;
		bounds.right = 640; //GetSystemMetrics(SM_CXSCREEN);
		bounds.bottom = 480; //GetSystemMetrics(SM_CYSCREEN);
		//AdjustWindowRectEx(&bounds, GetWindowLong(SDL_Window, GWL_STYLE), FALSE, 0);
		SetWindowPos(SDL_Window, HWND_TOPMOST,
			bounds.left, bounds.top, 
			bounds.right - bounds.left,
			bounds.bottom - bounds.top, SWP_NOCOPYBITS);
		//SDL_resizing = 0;
		ShowWindow(SDL_Window, SW_SHOW);
		SetForegroundWindow(SDL_Window);
		/*
		while ( GetForegroundWindow() != SDL_Window ) {
			SetForegroundWindow(SDL_Window);
			SDL_Delay(100);
		}
		/**/
	} else {
		RECT window_rect;

		window_rect.left = 0;
		window_rect.top = 0;
		window_rect.right = width;
		window_rect.bottom = height;
		AdjustWindowRect(&window_rect, style, FALSE);

		/* Resize the window */
		SetWindowPos(SDL_Window, HWND_TOP,
			0,
			0,
			window_rect.right - window_rect.left,
			window_rect.bottom - window_rect.top,
			SWP_NOCOPYBITS | SWP_NOMOVE);

		/* Show the window */
		ShowWindow(SDL_Window, SW_SHOW);

#if 1
		{
			/* Double check on the size of the window's client area */
			RECT temp_rect;

			GetClientRect(SDL_Window, &temp_rect);
			if ( (temp_rect.right - temp_rect.left) != width ) {
				SDL_SetError("width mismatch in SetVideoMode");
				return NULL;
			}
			if ( (temp_rect.bottom - temp_rect.top) != height ) {
				SDL_SetError("height mismatch in SetVideoMode");
				return NULL;
			}
		}
#endif
	}


	/*
	ShowWindow(SDL_Window, SW_SHOW);
	UpdateWindow(SDL_Window);
	/**/

	/*
	SetWindowPos(
		SDL_Window,
		0,
		0,
		0,
		width,
		height,
		SWP_NOMOVE | SWP_NOZORDER);
		/**/

	/* Set the appropriate window style */
	windowX = SDL_windowX;
	windowY = SDL_windowY;
	/*
	style = GetWindowLong(SDL_Window, GWL_STYLE);
	style &= ~(resizestyle | WS_MAXIMIZE);
	if ( (flags & SDL_FULLSCREEN) == SDL_FULLSCREEN ) {
		style &= ~windowstyle;
		style |= directstyle;
	} else {
		if ( flags & SDL_NOFRAME ) {
			style &= ~windowstyle;
			style |= directstyle;
		} else {
			style &= ~directstyle;
			style |= windowstyle;
			if ( flags & SDL_RESIZABLE ) {
				style |= resizestyle;
			}
		}
#if WS_MAXIMIZE
		if (IsZoomed(SDL_Window)) style |= WS_MAXIMIZE;
#endif
	}
	/**/
	/* DJM: Don't piss of anyone who has setup his own window */
	//if ( SDL_windowid == NULL )
	//	SetWindowLong(SDL_Window, GWL_STYLE, style);
	SDL_windowX = windowX;
	SDL_windowY = windowY;

	/* Set the display mode, if we are in fullscreen mode */
	if ( (flags & SDL_FULLSCREEN) == SDL_FULLSCREEN ) {
		/* FIXME, davidl: Implement fullscreen mode. */
	}

	/*
	 * Initialize Direct3D presentation parameters
	 */
	memset(&d3d_presentation_parameters, 0, sizeof(D3DPRESENT_PARAMETERS));

	/* Determine and set the pixel format */
	if ( (flags & SDL_FULLSCREEN) == SDL_FULLSCREEN ) {
		format = D3DFMT_R5G6B5;
	} else {
		format = D3DFMT_UNKNOWN;	/* FIXME, davidl: Look at 'vformat' for this info. */
	}

	/* Enable vsync? */
	if ( (g_ScreenFlags & SDLD3DSCREENFLAG_VSYNC) == SDLD3DSCREENFLAG_VSYNC ) {
		presentation_interval = D3DPRESENT_INTERVAL_DEFAULT;
	} else {
		presentation_interval = D3DPRESENT_INTERVAL_IMMEDIATE;
	}

	/* Create a single back buffer */
	d3d_presentation_parameters.BackBufferCount = 1;
	/* No multi-sampling */
	d3d_presentation_parameters.MultiSampleType = D3DMULTISAMPLE_NONE;
	d3d_presentation_parameters.MultiSampleQuality = 0;
	/* Throw away previous frames */
	d3d_presentation_parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;
	/* Set the main (and only) window */
	d3d_presentation_parameters.hDeviceWindow = SDL_Window;
	/* No flags to set */
	d3d_presentation_parameters.Flags = 0;
	/* Default refresh rate */
	d3d_presentation_parameters.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
	/* Default presentation rate */
	d3d_presentation_parameters.PresentationInterval = presentation_interval;
	/* Display format */
	d3d_presentation_parameters.BackBufferFormat = format;
	/* No depth/stencil buffer */
	d3d_presentation_parameters.EnableAutoDepthStencil = FALSE;

	if ( (flags & SDL_FULLSCREEN) == SDL_FULLSCREEN ) {
		d3d_presentation_parameters.Windowed = FALSE;
		d3d_presentation_parameters.BackBufferWidth = width;
		d3d_presentation_parameters.BackBufferHeight = height;
	} else {
		d3d_presentation_parameters.Windowed = TRUE;
	}

	/* Initialize the Direct3D device */
	VERIFY_FCALL(IDirect3D9_CreateDevice, (
		d3d9_object,
		/* Use the default adapter.  FIXME: allow SDL to specify this. */
		D3DADAPTER_DEFAULT,
		/* Use hardware acceleration rather than the software renderer. */
		SDL_D3D_DEVICE_TYPE,
		/* The window. */
		SDL_Window,
		/* Process vertices in software.  This is slower than in hardware,
		but will work on all graphics cards. */
		D3DCREATE_SOFTWARE_VERTEXPROCESSING,
		/* The D3DPRESENT_PARAMETERS structure, which contains additional
		information on the device to create. */
		&d3d_presentation_parameters,
		/* The following will point to the new device. */
		&d3d9_device),
		/* return NULL on error */
		return NULL);

	//ShowWindow(SDL_Window, SW_SHOW);
	//UpdateWindow(SDL_Window);

#if 0
	video = SDL_CreateRGBSurface(SDL_SWSURFACE, width, height, 32, 0x00FF0000, 0x0000FF00, 0x000000FF, 0xFF000000);
	return video;
#endif

	/* Create the display's corresponding SDL_Surface. */
	video = SDL_CreateRGBSurfaceFrom(
		/* The pixel data will be filled in later. */
		NULL,
		/* Dimensions */
		width, height,
		/* for now, only 32-bit, ARGB color is supported */
		32,
		/* pitch = (width * pixel_size_in_bytes) + extra_alignment_bytes */
		(width * 4),
		/* use 32-bit, ARGB color, with 1 byte per component */
		0x00FF0000,	/* red */
		0x0000FF00, /* green */
		0x000000FF,	/* blue */
		0xFF000000	/* alpha */
		);

	/* Create the display surface's private data structure */
	surfaceData = video->hwdata = (struct private_hwdata *) malloc(sizeof(struct private_hwdata));
	if ( video->hwdata == NULL ) {
		SDL_OutOfMemory();
		return NULL;
	}
	memset(video->hwdata, 0, sizeof(struct private_hwdata));

	/* Set flags in the display surface */
	video->flags &= ~SDL_SWSURFACE;
	video->flags |= SDL_HWSURFACE | SDL_DOUBLEBUF;	/* FIXME, davidl: SDL_DOUBLEBUF is set in all cases. */
	surfaceData->flags = D3DSURFACEFLAG_IS_DISPLAY;

	/* Retrieve the back buffer */
	VERIFY_FCALL(IDirect3DDevice9_GetBackBuffer, (
		d3d9_device,
		0,
		0,
		D3DBACKBUFFER_TYPE_MONO,
		&surfaceData->surface),
		return NULL);

	/* Store pointer to back buffer in the display surface */
	//D3DPDATA(video)->pixel_data = (void *) backBuffer;

	/* Update info on video capabilities */
	DIRECT3D_UpdateVideoInfo(this);

	/* All done! */
	return video;
}

static int DIRECT3D_AllocHWSurface(_THIS, SDL_Surface *surface)
{
	/* Create private surface data */
	surface->hwdata = (struct private_hwdata *) malloc(sizeof(struct private_hwdata));
	if ( surface->hwdata == NULL ) {
		SDL_OutOfMemory();
		return -1;
	}
	memset(surface->hwdata, 0, sizeof(struct private_hwdata));

	/* Set blend mode */
	surface->hwdata->blend_mode = SDLD3DBLEND_NORMAL;

	/* Create texture(s) */
	VERIFY_FCALL(TextureGrid_create, (
		this,
		/* Surface dimensions */
		surface->w, surface->h,
		/* Maximum grid-cell dimensions */
		64, 64, //256, 256, //128, 128, //64, 64,
		/* Output */
		&surface->hwdata->texture_grid),
		/* return -1 on error */
		return -1);

	/* Set flags */
	surface->flags |= SDL_HWSURFACE;

	/* Set pixel data field to NULL.  It will be filled in if and when the
	   surface is locked. */
	surface->pixels = NULL;

	/* All done! */
	return 0;
}

static int DIRECT3D_CheckHWBlit(_THIS, SDL_Surface *src, SDL_Surface *dst)
{
	/* Both 'src' and 'dst' must be hardware surfaces in order for an
	   accelerated blit to occur. */
	if ( !(src->flags & SDL_HWSURFACE) || (dst->flags & SDL_SWSURFACE) )
		return 0;

	/* For now, the destination surface MUST be the display/screen. */
	if ( (dst->hwdata->flags & D3DSURFACEFLAG_IS_DISPLAY) == 0 )
		return 0;

	/* For now, assume that all blits between hardware surfaces can be
	   accelerated. */
	src->flags |= SDL_HWACCEL;

	/* Inform SDL of the blit function. */
	src->map->hw_blit = DIRECT3D_HWAccelBlit;

	/* Return 1, which means that an accelerated blit is possible. */
	return 0;
}

static int DIRECT3D_HWAccelBlit(SDL_Surface *src, SDL_Rect *srcrect, SDL_Surface *dst, SDL_Rect *dstrect)
{
	HRESULT result;
	Uint8 src_surface_alpha;

	if (srcrect->w != dstrect->w)
		return -1;

	if (srcrect->h != dstrect->h)
		return -1;

	result = set_device_states(current_video, SDLD3D_DEVMODE_DRAWTEXTURES, src->hwdata->blend_mode);
	if ( FAILED(result) )
		return -1;

	result = begin_rendering(current_video);
	if ( FAILED(result) )
		return -1;

	src_surface_alpha = (src->hwdata->blend_mode == SDLD3DBLEND_MIXEDALPHA) ? src->format->alpha : SDL_ALPHA_OPAQUE;

	VERIFY_FCALL(TextureGrid_draw, (
		src->hwdata->texture_grid,
		current_video->hidden->__d3d9_device,
		srcrect,
		dstrect,
		src_surface_alpha),
		return -1);

	result = stop_rendering(current_video);
	if ( FAILED(result) )
		return -1;

	return 0;
}

static HRESULT render_color(_THIS, SDL_Rect *dest_rect, DWORD color)
{
	/* Prevent rounding errors */
	float realX = (float)dest_rect->x + 0.5f;
	float realY = (float)dest_rect->y + 0.5f;
	int width = dest_rect->w;
	int height = dest_rect->h;

	/* Define a quad by using a triangle strip */
	struct textured_vertex vertices[] = {
		{ realX,			realY,			0,	1,	color,	0,	0 },
		{ realX + width,	realY,			0,	1,	color,	1,	0 },
		{ realX,			realY + height,	0,	1,	color,	0,	1 },
		{ realX + width,	realY + height,	0,	1,	color,	1,	1 }
	};

	/* Render the quad */
	VERIFY_FCALL(IDirect3DDevice9_DrawPrimitiveUP, (
		/* Device to render with */
		d3d9_device,
		/* PrimitiveType */
		D3DPT_TRIANGLESTRIP,
		/* PrimitiveCount (number of triangles) */
		2,
		/* pVertexStreamZeroData (triangle data) */
		vertices,
		/* VertexStreamZeroStride */
		sizeof(struct textured_vertex)),
		/* return generated error code on failure */
		return _fcall_hresult);

	/* All done! */
	return S_OK;
}

static int DIRECT3D_FillHWRect(_THIS, SDL_Surface *dst, SDL_Rect *dstrect, Uint32 color)
{
#if 0
	VERIFY_FCALL(set_device_states, (this, SDLD3D_DEVMODE_DRAWSOLIDCOLORS, 0),
		return -1);

	VERIFY_FCALL(begin_rendering, (this),
		return -1);

	VERIFY_FCALL(render_color, (this, dstrect, color),
		return -1);

	VERIFY_FCALL(stop_rendering, (this),
		return -1);
#elif 0
	VERIFY_FCALL(IDirect3DDevice9_Clear, (
		d3d9_device,
		0,
		NULL,
		D3DCLEAR_TARGET,
		color,
		1.0f,
		0),
		return -1);
#elif 1
	RECT rect = {
		dstrect->x,
		dstrect->y,
		dstrect->x + dstrect->w,
		dstrect->y + dstrect->h
	};
	IDirect3DSurface9 *surface = dst->hwdata->surface;

	if ( surface == NULL ) {
		SDL_SetError("SDL's direct3d driver can only colorfill to the screen.");
		return -1;
	}

	VERIFY_FCALL(IDirect3DDevice9_ColorFill, (
		/* Device to render with */
		d3d9_device,
		/* pSurface */
		surface,
		/* pRect */
		&rect,
		/* color */
		color),
		/* return -1 on error */
		return -1);
#endif

	/* All done! */
	return 0;
}

static int DIRECT3D_SetHWColorKey(_THIS, SDL_Surface *surface, Uint32 key)
{
	return -1;
}

static int DIRECT3D_SetHWAlpha(_THIS, SDL_Surface *surface, Uint8 alpha)
{
	return -1;
}

static int DIRECT3D_LockHWSurface(_THIS, SDL_Surface *surface)
{
	if (surface->hwdata->flags & D3DSURFACEFLAG_LOCKED)
	{
		SDL_SetError("The surface has already been locked.");
		return -1;
	}

	if (surface->hwdata->flags & D3DSURFACEFLAG_IS_DISPLAY)
	{
		SDL_SetError("A direct3d backbuffer cannot be accessed/locked, yet.");
		return -1;

#if 0
		result = IDirect3DSurface9_LockRect(
			/* The surface to lock */
			surface->hwdata->surface,
			/* Locked rect structure, which will contain the pointer to the pixel buffer. */
			&lockedRect,
			/* Area to lock (lock everything) */
			NULL,
			/* Flags */
			0);

		if ( FAILED(result) ) {
			SDL_SetError("Call to IDirect3DSurface9_LockRect failed (in SDL_LockSurface -> DIRECT3D_LockSurface): %s", DescribeLastWinError());
			return -1;
		}
#endif
	}
	else
	{
		if ( surface->hwdata->surface_to_upload != NULL ) {
			/* A surface is already waiting to be uploaded.  Use that one. */
			return 0;
		}

		surface->hwdata->surface_to_upload = SDL_CreateRGBSurface(
			/* Create a surface in system memory */
			SDL_SWSURFACE,
			/* Create a surface as similar to the one in hardware */
			surface->w, surface->h,
			surface->format->BitsPerPixel,
			surface->format->Rmask,
			surface->format->Gmask,
			surface->format->Bmask,
			surface->format->Amask);
		if ( surface->hwdata->surface_to_upload == NULL ) {
			/* The uploadable surface couldn't be created! */
			return -1;
		}

		/* Make sure that blits get directed to surface_to_upload */
		//surface->hwdata->saved_pixels = surface->pixels;
		surface->hwdata->saved_pitch = surface->pitch;
		surface->hwdata->saved_offset = surface->offset;
		surface->pixels = surface->hwdata->surface_to_upload->pixels;
		surface->pitch = surface->hwdata->surface_to_upload->pitch;
		surface->offset = surface->hwdata->surface_to_upload->offset;
		return 0;
	}

	/* Set flags */
	surface->hwdata->flags |= D3DSURFACEFLAG_LOCKED;
	
	/* All done! */
	return 0;
}

static void DIRECT3D_UnlockHWSurface(_THIS, SDL_Surface *surface)
{
	//HRESULT result;

	/*
	if ((surface->hwdata->flags & D3DSURFACEFLAG_LOCKED) == 0)
	{
		return;
	}
	/**/

	/* Restore old member variables */
	surface->pixels = NULL;
	surface->pitch = surface->hwdata->saved_pitch;
	surface->offset = surface->hwdata->saved_offset;

	/* Copy new data to the texture grid */
	VERIFY_FCALL(TextureGrid_copyPixelsFromSurface, (
		/* texture grid */
		surface->hwdata->texture_grid,
		/* source pixel data */
		surface->hwdata->surface_to_upload),
		/* return nothing on error */
		return);

	/* Destroy surface_to_upload, as it is no longer needed. */
	SDL_FreeSurface(surface->hwdata->surface_to_upload);
	surface->hwdata->surface_to_upload = NULL;
}

static int DIRECT3D_FlipHWSurface(_THIS, SDL_Surface *surface)
{
	DIRECT3D_UpdateRects(this, 0, NULL);
	return 0;

	//return -1;
}

static void DIRECT3D_FreeHWSurface(_THIS, SDL_Surface *surface)
{
	if ( surface->hwdata ) {
		if ( surface->hwdata->texture_grid ) {
			TextureGrid_destroy(surface->hwdata->texture_grid);
		}

		free(surface->hwdata);
	}
}

static HRESULT is_device_ready(_THIS)
{
	VERIFY_FCALL(IDirect3DDevice9_TestCooperativeLevel, (d3d9_device),
		return _fcall_hresult);

	return S_OK;
}

static HRESULT set_device_state_to_drawtextures(_THIS, SDLD3D_BlendMode blend_mode)
{
	/* Enable alpha blending */
	VERIFY_FCALL(IDirect3DDevice9_SetRenderState, (d3d9_device, D3DRS_ALPHABLENDENABLE, TRUE),
		return _fcall_hresult);
	VERIFY_FCALL(IDirect3DDevice9_SetRenderState, (d3d9_device, D3DRS_SRCBLEND, D3DBLEND_SRCALPHA),
		return _fcall_hresult);
	VERIFY_FCALL(IDirect3DDevice9_SetRenderState, (d3d9_device, D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA),
		return _fcall_hresult);

	switch (blend_mode)
	{
	case SDLD3DBLEND_NORMAL:
		VERIFY_FCALL(IDirect3DDevice9_SetTextureStageState, (d3d9_device, 0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1),
			return _fcall_hresult);
		VERIFY_FCALL(IDirect3DDevice9_SetTextureStageState, (d3d9_device, 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE),
			return _fcall_hresult);
		VERIFY_FCALL(IDirect3DDevice9_SetTextureStageState, (d3d9_device, 0, D3DTSS_ALPHAARG2, D3DTA_TEXTURE),
			return _fcall_hresult);
		break;

	case SDLD3DBLEND_MIXEDALPHA:
		VERIFY_FCALL(IDirect3DDevice9_SetTextureStageState, (d3d9_device, 0, D3DTSS_ALPHAOP, D3DTOP_MODULATE),
			return _fcall_hresult);
		VERIFY_FCALL(IDirect3DDevice9_SetTextureStageState, (d3d9_device, 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE),
			return _fcall_hresult);
		VERIFY_FCALL(IDirect3DDevice9_SetTextureStageState, (d3d9_device, 0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE),
			return _fcall_hresult);
		break;
	}

	VERIFY_FCALL(IDirect3DDevice9_SetTextureStageState, (d3d9_device, 0, D3DTSS_COLOROP, D3DTOP_SELECTARG1),
		return _fcall_hresult);
	VERIFY_FCALL(IDirect3DDevice9_SetTextureStageState, (d3d9_device, 0, D3DTSS_COLORARG1, D3DTA_TEXTURE),
		return _fcall_hresult);
	VERIFY_FCALL(IDirect3DDevice9_SetTextureStageState, (d3d9_device, 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE),
		return _fcall_hresult);

	return S_OK;
}

static HRESULT set_device_state_to_drawcolidcolors(_THIS)
{
	/* Enable alpha blending */
	VERIFY_FCALL(IDirect3DDevice9_SetRenderState, (d3d9_device, D3DRS_ALPHABLENDENABLE, TRUE),
		return _fcall_hresult);
	VERIFY_FCALL(IDirect3DDevice9_SetRenderState, (d3d9_device, D3DRS_SRCBLEND, D3DBLEND_SRCALPHA),
		return _fcall_hresult);
	VERIFY_FCALL(IDirect3DDevice9_SetRenderState, (d3d9_device, D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA),
		return _fcall_hresult);

	VERIFY_FCALL(IDirect3DDevice9_SetTextureStageState, (d3d9_device, 0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1),
		return _fcall_hresult);
	VERIFY_FCALL(IDirect3DDevice9_SetTextureStageState, (d3d9_device, 0, D3DTSS_ALPHAARG1, D3DTA_DIFFUSE),
		return _fcall_hresult);
	VERIFY_FCALL(IDirect3DDevice9_SetTextureStageState, (d3d9_device, 0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE),
		return _fcall_hresult);

	VERIFY_FCALL(IDirect3DDevice9_SetTextureStageState, (d3d9_device, 0, D3DTSS_COLOROP, D3DTOP_SELECTARG1),
		return _fcall_hresult);
	VERIFY_FCALL(IDirect3DDevice9_SetTextureStageState, (d3d9_device, 0, D3DTSS_COLORARG1, D3DTA_DIFFUSE),
		return _fcall_hresult);
	VERIFY_FCALL(IDirect3DDevice9_SetTextureStageState, (d3d9_device, 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE),
		return _fcall_hresult);

	return S_OK;
}


static HRESULT set_device_states(_THIS, int mode, SDLD3D_BlendMode blend_mode)
{
	switch (mode)
	{
	case SDLD3D_DEVMODE_DRAWTEXTURES:
		return set_device_state_to_drawtextures(this, blend_mode);

	case SDLD3D_DEVMODE_DRAWSOLIDCOLORS:
		//return set_device_state_to_drawtextures(this);
		return set_device_state_to_drawcolidcolors(this);

	default:
		SDL_SetError("Unknown 'mode' passed to set_device_states (in SDL_direct3dvideo.c).");
		return E_FAIL;
	}

	return S_OK;
}

static HRESULT begin_rendering(_THIS)
{
	VERIFY_FCALL(IDirect3DDevice9_BeginScene, (d3d9_device),
		return _fcall_hresult);

	VERIFY_FCALL(IDirect3DDevice9_SetFVF, (d3d9_device, tri_fvf),
		return _fcall_hresult);

	//set_device_states(this, SDLD3D_DEVMODE_DRAWTEXTURES);

	return S_OK;
}

static HRESULT stop_rendering(_THIS)
{
	VERIFY_FCALL(IDirect3DDevice9_EndScene, (d3d9_device),
		return _fcall_hresult);

	return S_OK;
}

static HRESULT present_data(_THIS)
{
	/* Present the contents of the back buffer to the user. */
	VERIFY_FCALL(IDirect3DDevice9_Present, (
		d3d9_device,
		NULL,
		NULL,
		NULL,
		NULL),
		/* return failed hresult on error */
		return _fcall_hresult);

	return S_OK;
}

void DIRECT3D_UpdateRects(_THIS, int numrects, SDL_Rect *rects)
{
	if ( FAILED(is_device_ready(this)) ) {
		exit(-1);
	}

	if ( FAILED(present_data(this)) ) {
		exit(-1);
	}
}

int DIRECT3D_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors)
{
	return 0;
}

static int DIRECT3D_SetGammaRamp(_THIS, Uint16 *ramp)
{
	return -1;
}

static int DIRECT3D_GetGammaRamp(_THIS, Uint16 *ramp)
{
	return -1;
}

static void FlushMessageQueue()
{
	MSG  msg;
	while ( PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) ) {
		if ( msg.message == WM_QUIT ) break;
		TranslateMessage( &msg );
		DispatchMessage( &msg );
	}
}

void DIRECT3D_VideoQuit(_THIS)
{
	//struct private_hwdata *_hwdata = current_video->hwdata;
	if ( SDL_VideoSurface != NULL ) {
		struct private_hwdata *_hwdata = (struct private_hwdata *) SDL_VideoSurface->hwdata;
		IDirect3DSurface9 *backBuffer = _hwdata->surface;
		IDirect3DSurface9_Release(backBuffer);
	}


	if ( SDL_Window ) {
		DIRECT3D_DestroyWindow(this);
		FlushMessageQueue();
	}
}

void DIRECT3D_RealizePalette(_THIS)
{
}

void DIRECT3D_PaletteChanged(_THIS, HWND window)
{
}

void DIRECT3D_WinPAINT(_THIS, HDC hdc)
{
}

LONG DIRECT3D_HandleMessage(_THIS, HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	return DIB_HandleMessage(this, hwnd, msg, wParam, lParam);
	//return DefWindowProc(hwnd, msg, wParam, lParam);
}

